from wr_profiles import Profile, Property


class WarehouseProfile(Profile):
    profile_root = 'warehouse'

    host = Property('host', default='localhost')
    username = Property('username')
    password = Property('password')



import pytest

from tests.warehouse_profile import WarehouseProfile as WP
from wr_profiles import Profile

case_matrix = [
    [
        (None, None, True),
        {
            'parent_profile': None,
        },
    ],
    [
        (None, None, False),
        {
            'parent_profile': None,
        },
    ],
    [
        ('staging', None, True),
        {
            'parent_profile': None,
        },
    ],
    [
        ('staging', None, False),
        {
            'parent_profile': None,
        },
    ],
    [
        (None, 'production', True),
        {
            'parent_profile': ('production', None, True),
        },
    ],
    [
        (None, 'production', False),
        {
            'parent_profile': ('production', None, False),
        },
    ],
    [
        ('staging', 'production', True),
        {
            'parent_profile': ('production', None, True),
        },
    ],
    [
        ('staging', 'production', False),
        {
            'parent_profile': ('production', None, False),
        },
    ],
]


@pytest.mark.parametrize('case', case_matrix)
def test_parent_profile_initialisation(case):
    profile_name, parent_profile_name, is_live = case[0]
    profile = WP(name=profile_name, parent_name=parent_profile_name, is_live=is_live)
    if case[1]['parent_profile'] is None:
        assert profile.parent_profile is None
    else:
        parent_profile_sig = (
            profile.parent_profile.profile_name,
            profile.parent_profile.parent_profile_name,
            profile.parent_profile.is_live,
        )
        assert parent_profile_sig == case[1]['parent_profile']


def test_profile_name(monkeypatch):
    monkeypatch.setenv('WAREHOUSE_PROFILE', 'production')
    assert WP().profile_name == 'production'
    assert WP(name='production').profile_name == 'production'
    assert WP(is_live=False).profile_name is None
    assert WP(parent_name='staging').profile_name == 'production'
    assert WP(name='staging', is_live=False).profile_name == 'staging'


def test_parent_profile_name(monkeypatch):
    # this should be ignored by all because the base profile cannot have a parent profile
    monkeypatch.setenv('WAREHOUSE_PARENT_PROFILE', 'illegal')

    monkeypatch.setenv('WAREHOUSE_STAGING_PARENT_PROFILE', 'production')

    assert WP().parent_profile_name is None
    assert WP().parent_profile is None

    assert WP(name='production').parent_profile_name is None
    assert WP(name='production').parent_profile is None

    assert WP(name='production', is_live=False).parent_profile_name is None
    assert WP(name='production', is_live=False).parent_profile is None

    assert WP(name='staging').parent_profile_name == 'production'
    assert WP(name='staging').parent_profile.profile_name == 'production'

    assert WP(name='staging', is_live=False).parent_profile_name is None
    assert WP(name='staging', is_live=False).parent_profile is None


def test_loaders():
    assert WP().loader is Profile._profile_loaders['live']
    assert WP(is_live=False).loader is Profile._profile_loaders['frozen']


def test_to_dict():
    assert WP().to_dict() == {WP.host: 'localhost'}
    assert WP(is_live=False).to_dict() == {WP.host: 'localhost'}



import os

import pytest


@pytest.fixture(autouse=True)
def reset_warehouse_profile_envvars(monkeypatch):
    """
    Reset any environment variables that could affect the WarehouseProfile used in tests.
    """
    for k in list(os.environ):
        if k.startswith('WAREHOUSE_'):
            del os.environ[k]
            monkeypatch.delenv(k, raising=False)



import pytest

from tests.warehouse_profile import WarehouseProfile
from wr_profiles import Property


def test_ultimate_live_profile():
    profile = WarehouseProfile()
    assert profile.profile_root == 'warehouse'

    assert profile._const_name is None
    assert profile._const_parent_name is None
    assert profile.is_live is True

    assert not profile.profile_name
    assert not profile.parent_profile_name
    assert not profile.parent_profile

    assert profile.is_active


def test_concrete_frozen_profile(monkeypatch):
    monkeypatch.setenv('WAREHOUSE_PROFILE', 'something_else')  # ignored
    monkeypatch.setenv('WAREHOUSE_STAGING_PARENT_PROFILE', 'not_production')  # ignored
    monkeypatch.setenv('WAREHOUSE_STAGING_USERNAME', 'staging-username')
    monkeypatch.setenv('WAREHOUSE_PRODUCTION_PASSWORD', 'production-password')

    live = WarehouseProfile(name='staging', parent_name='production', is_live=True)
    frozen = WarehouseProfile(name='staging', parent_name='production', is_live=False)

    assert live.username == 'staging-username'
    with pytest.raises(Property.MissingValue):
        assert frozen.username

    assert live.password == 'production-password'
    with pytest.raises(Property.MissingValue):
        assert frozen.password

    assert live.host == 'localhost'
    assert frozen.host == 'localhost'

    frozen.load()

    assert live.host == frozen.host
    assert live.username == frozen.username
    assert live.password == frozen.password



from pathlib import Path

import pytest

from tests.warehouse_profile import WarehouseProfile
from wr_profiles import Profile, Property


class UploaderProfile(Profile):
    profile_root = 'uploader'

    root_dir = Property('root_dir', default=Path('/'))
    user = Property('user')

    enabled = Property('enabled', default=False)

    @enabled.deserializer
    def enabled(self, value):
        if isinstance(value, bool):
            return value
        return value and value[0].lower()[0] in ('y', 't', '1')

    @enabled.serializer
    def enabled(self, value):
        if value:
            return str(value).lower()[0]
        return '0'


uploader_profile = UploaderProfile()


def test_all(monkeypatch):
    assert isinstance(UploaderProfile.root_dir, Property)
    assert UploaderProfile.root_dir.name == 'root_dir'

    assert uploader_profile.root_dir == Path('/')

    with pytest.raises(Property.MissingValue):
        assert not uploader_profile.user

    monkeypatch.setenv('UPLOADER_USER', 'uploader')
    assert uploader_profile.user == 'uploader'

    assert uploader_profile.enabled is False
    assert uploader_profile.to_envvars() == {
        'UPLOADER_USER': 'uploader',
        'UPLOADER_ROOT_DIR': '/',
        'UPLOADER_ENABLED': '0',
    }

    monkeypatch.setenv('UPLOADER_ENABLED', 'yes')
    assert uploader_profile.enabled is True
    assert uploader_profile.to_envvars() == {
        'UPLOADER_USER': 'uploader',
        'UPLOADER_ROOT_DIR': '/',
        'UPLOADER_ENABLED': 't',
    }


def test_integers_are_converted_to_strings():
    staging = WarehouseProfile(name='staging')
    staging.username = 5555
    assert staging.to_envvars() == {
        'WAREHOUSE_STAGING_HOST': 'localhost',
        'WAREHOUSE_STAGING_USERNAME': '5555',
    }


@pytest.mark.parametrize('is_live', [
    True,
    False,
])
def test_to_envvars_includes_parent_profile_setting(is_live):
    staging = WarehouseProfile(name='staging', parent_name='production', is_live=is_live, values={
        'username': 'hard-coded-staging-username',
    })

    assert staging.to_envvars() == {
        'WAREHOUSE_STAGING_PARENT_PROFILE': 'production',
        'WAREHOUSE_STAGING_HOST': 'localhost',
        'WAREHOUSE_STAGING_USERNAME': 'hard-coded-staging-username',
    }



from typing import Any

from .not_set import NotSet


class Property:
    """
    Represents a profile property -- a value backed by an environment variable.
    The exact environment variable depends on what profile the property belongs to.
    """

    class MissingValue(Exception):
        """
        Raised when the value is requested for an existing property
        """
        def __init__(self, prop):
            self.prop = prop

        def __repr__(self):
            return '<{} {!r}>'.format(self.__class__.__name__, self.prop.name)

    def __init__(self, name=None, default: Any=NotSet, deserializer=None, serializer=None):
        self.name = name
        self.default = default
        self._deserializer = deserializer
        self._serializer = serializer

    def __hash__(self):
        return hash((self.__class__, self.name))

    def __set_name__(self, owner, name):
        # Only works in Python 3.6+
        # On Python 3.5 you need to manually pass the name.
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.get_prop_value(self)

    def __set__(self, instance, value):
        instance.set_prop_value(self, value)

    def __str__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.name)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.name)

    def get_envvar(self, profile):
        assert self.name
        return '{}{}'.format(profile.envvar_prefix, self.name.upper())

    @property
    def has_default(self):
        """
        Returns True if this property has a default value set in its declaration.
        """
        return self.default is not NotSet

    def serializer(self, func):
        """
        Register a serializer for this profile property.
        Serializer is a function that takes property value and returns a string.
        The function name must be either `<lambda>` or `{prop_name}`
        """
        if callable(func):
            if not (func.__name__ == '<lambda>' or func.__name__ == self.name):
                raise RuntimeError('Invalid {!r} property serializer name {!r} -- should be called {!r}'.format(
                    self.name, func.__name__, self.name
                ))
        else:
            assert func is None
        self._serializer = func
        return self

    def deserializer(self, func):
        """
        Register a deserializer for this profile property.
        The function name must be either `<lambda>` or `{prop_name}`.
        """
        if callable(func):
            if not (func.__name__ == '<lambda>' or func.__name__ == self.name):
                raise RuntimeError('Invalid {!r} property deserializer name {!r} -- should be called {!r}'.format(
                    self.name, func.__name__, self.name
                ))
        else:
            assert func is None
        self._deserializer = func
        return self

    def from_str(self, profile, value):
        if self._deserializer is not None:
            return self._deserializer(profile, value)
        else:
            return value

    def to_str(self, profile, value):
        if self._serializer is not None:
            return self._serializer(profile, value)
        else:
            return str(value)

    def missing_value(self):
        return self.MissingValue(self)


class AttributesList:
    def __init__(self, attr_base_cls):
        self._attr_base_cls = attr_base_cls

    def __get__(self, instance, owner):
        if instance is None:
            return self

        for name in dir(owner):
            if name.startswith('__'):
                continue
            attr = getattr(owner, name)
            if isinstance(attr, self._attr_base_cls):
                yield attr



import os
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, Union

from .not_set import NotSet
from .props import AttributesList, Property

PROFILE_NAME_COMPONENT_REGEX = re.compile(r'^[a-z]([\d\w]*[a-z0-9])?$')


class ProfileLoader(ABC):
    """
    Base class for profile loaders.
    """

    @abstractmethod
    def load(self, profile):
        pass

    @abstractmethod
    def get_prop_value(self, profile: 'Profile', prop: Union[str, Property], default: Any=NotSet) -> Any:
        pass

    @abstractmethod
    def set_prop_value(self, profile: 'Profile', prop: Union[str, Property], value: Any):
        pass

    def to_dict(self, profile: 'Profile') -> Dict[Property, Any]:
        values = {}
        for prop in profile.props:
            try:
                values[prop] = profile.get_prop_value(prop)
            except Property.MissingValue:
                pass
        return values


class LiveProfileLoader(ProfileLoader):
    def set_prop_value(self, profile: 'Profile', prop: Union[str, Property], value: Any):
        if isinstance(prop, str):
            prop = profile.get_prop(prop)
        os.environ[prop.get_envvar(profile)] = prop.to_str(profile, value)

    def get_prop_value(self, profile: 'Profile', prop: Union[str, Property], default: Any = NotSet) -> Any:
        if isinstance(prop, str):
            prop = profile.get_prop(prop)

        for check_profile in profile.get_profile_tree():
            if prop.name in check_profile._const_values:
                return check_profile._const_values[prop.name]

        for check_profile in profile.get_profile_tree():
            prop_envvar = prop.get_envvar(check_profile)
            if prop_envvar in os.environ:
                return prop.from_str(check_profile, os.environ[prop_envvar])

        if default is not NotSet:
            return default

        if prop.has_default:
            return prop.default

        raise prop.missing_value()

    def load(self, profile):
        # Nothing to do -- live profile does not need to be reloaded.
        pass


class FrozenProfileLoader(ProfileLoader):
    def set_prop_value(self, profile: 'Profile', prop: Union[str, Property], value: Any):
        if isinstance(prop, str):
            prop = profile.get_prop(prop)
        profile._const_values[prop.name] = value

    def get_prop_value(self, profile: 'Profile', prop: Union[str, Property], default: Any = NotSet) -> Any:
        if isinstance(prop, str):
            prop = profile.get_prop(prop)

        for check_profile in profile.get_profile_tree():
            if prop.name in check_profile._const_values:
                return check_profile._const_values[prop.name]

        if default is not NotSet:
            return default

        if prop.has_default:
            return prop.default

        raise prop.missing_value()

    def load(self, profile):
        # Create a live clone of itself and load all props.
        live_clone = profile.__class__(
            name=profile.profile_name,
            parent_name=profile.parent_profile_name,
            is_live=True,
            values=profile._const_values,
        )

        values = {}
        for prop in profile.props:
            try:
                values[prop.name] = live_clone.get_prop_value(prop.name)
            except Property.MissingValue:
                continue

        profile._const_values = values


class Profile:
    """
    Represents a set of configuration values backed by environment variables.
    """

    profile_root = None

    props = AttributesList(Property)

    # shared loaders
    _profile_loaders = {}  # type: Dict[str, ProfileLoader]

    def __init__(self, *, name=None, parent_name=None, is_live=True, values=None, **kwargs):
        self._const_name = name
        self._const_parent_name = parent_name
        self._const_is_live = is_live

        self._const_values = {}
        if values is not None:
            self._const_values.update(values)

        if kwargs:
            raise ValueError(kwargs)

        if not self.profile_root:
            raise ValueError('{}.profile_root is required'.format(self.__class__.__name__))

        if not PROFILE_NAME_COMPONENT_REGEX.match(self.profile_root):
            raise ValueError('{}.profile_root {!r} is invalid'.format(self.__class__.__name__, self.profile_root))

    @classmethod
    def get_instance(cls, name=None, parent_name=None, is_live=False, values=None):
        """
        Get a loaded frozen instance of a specific profile.
        """
        instance = cls(name=name, parent_name=parent_name, is_live=is_live, values=values)
        instance.load()
        return instance

    @property
    def envvar_prefix(self):
        if self.profile_name:
            return '{}_{}_'.format(self.profile_root, self.profile_name).upper()
        return '{}_'.format(self.profile_root).upper()

    @property
    def parent_profile_name(self):
        if self._const_parent_name:
            return self._const_parent_name
        elif not self.is_live:
            return None
        elif self.profile_name:
            return os.environ.get('{}PARENT_PROFILE'.format(self.envvar_prefix), None)
        else:
            return None

    @property
    def profile_name(self):
        if self._const_name:
            return self._const_name
        elif not self.is_live:
            return None
        else:
            return self.active_profile_name

    @property
    def active_profile_name(self):
        return os.environ.get('{}_PROFILE'.format(self.profile_root).upper(), None) or None

    @active_profile_name.setter
    def active_profile_name(self, value):
        if value is None:
            value = ''
        os.environ['{}_PROFILE'.format(self.profile_root).upper()] = value

    @property
    def is_live(self):
        return self._const_is_live

    @property
    def is_active(self):
        return self.profile_name == self.active_profile_name

    @property
    def parent_profile(self):
        profile_name = self.parent_profile_name
        if profile_name is None:
            return None
        else:
            return self.__class__(name=self.parent_profile_name, parent_name=None, is_live=self.is_live)

    def get_prop(self, prop_name):
        prop = getattr(self.__class__, prop_name, None)
        if prop is None or not isinstance(prop, Property):
            raise KeyError(prop_name)
        return prop

    def get_profile_tree(self):
        yield self
        parent_profile = self.parent_profile
        while parent_profile:
            yield parent_profile
            parent_profile = parent_profile.parent_profile

    @property
    def loader(self):
        if self.is_live:
            if 'live' not in self._profile_loaders:
                self._profile_loaders['live'] = LiveProfileLoader()
            return self._profile_loaders['live']
        else:
            if 'frozen' not in self._profile_loaders:
                self._profile_loaders['frozen'] = FrozenProfileLoader()
            return self._profile_loaders['frozen']

    def get_prop_value(self, prop: Union[str, Property], default=NotSet):
        return self.loader.get_prop_value(self, prop, default=default)

    def set_prop_value(self, prop: Union[str, Property], value: Any):
        self.loader.set_prop_value(self, prop, value)

    def load(self):
        self.loader.load(self)

    def to_dict(self) -> Dict[Property, Any]:
        return self.loader.to_dict(self)

    def to_envvars(self):
        """
        Export property values to a dictionary with environment variable names as keys.
        """
        export = {}
        for prop, prop_value in self.to_dict().items():
            export[prop.get_envvar(self)] = prop.to_str(self, prop_value)
        if self.parent_profile_name:
            export['{}PARENT_PROFILE'.format(self.envvar_prefix).upper()] = self.parent_profile_name
        return export

    def activate(self, profile_name=NotSet):
        """
        Sets <PROFILE_ROOT>_PROFILE environment variable to the name of the current profile.
        """
        if profile_name is NotSet:
            profile_name = self.profile_name
        self.active_profile_name = profile_name



from tests.warehouse_profile import WarehouseProfile


def test_default_profile_is_active_and_live():
    profile = WarehouseProfile()
    assert profile.profile_name is None
    assert profile.is_active
    assert profile.is_live


def test_custom_profile_is_frozen_but_could_be_active(monkeypatch):
    monkeypatch.setenv('WAREHOUSE_PROFILE', 'staging')

    profile = WarehouseProfile.get_instance('staging')
    assert profile.profile_name == 'staging'
    assert profile.is_active

    monkeypatch.setenv('WAREHOUSE_PROFILE', '')
    assert not profile.is_active

    monkeypatch.setenv('WAREHOUSE_PROFILE', 'staging')
    assert profile.is_active



import pytest

from tests.warehouse_profile import WarehouseProfile


@pytest.mark.parametrize('profile_name', [
    'staging',
    None,
])
def test_activates_live_profile(profile_name):
    wp = WarehouseProfile()

    instance = WarehouseProfile(name=profile_name)
    instance.activate()

    assert wp.active_profile_name == profile_name
    assert instance.is_active


@pytest.mark.parametrize('profile_name', [
    'staging',
    None,
])
def test_activates_frozen_profile(profile_name):
    wp = WarehouseProfile()

    instance = WarehouseProfile(name=profile_name, is_live=False)
    instance.activate()

    assert instance.is_active
    assert wp.active_profile_name == profile_name


def test_activates_named_profile():
    wp = WarehouseProfile()

    wp.activate('staging')
    assert wp.active_profile_name == 'staging'

    wp.activate('production')
    assert wp.active_profile_name == 'production'

    wp.activate(None)
    assert wp.active_profile_name is None



import pytest

from tests.warehouse_profile import WarehouseProfile


@pytest.mark.parametrize('profile_name', [
    'staging',
    None,
])
def test_const_values_set_on_frozen_profile_instance(profile_name, monkeypatch):
    wp = WarehouseProfile.get_instance(name=profile_name, values={
        'host': 'localhost.test',
        'username': 'test',
    })

    assert not wp.is_live
    assert wp.host == 'localhost.test'
    assert wp.username == 'test'
    assert wp.to_dict() == {
        WarehouseProfile.host: 'localhost.test',
        WarehouseProfile.username: 'test',
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()


setup(
    name='wr-profiles',
    version=read('wr_profiles/__init__.py').split('\n')[0].split('=', 1)[1].strip().strip("'"),
    author='Jazeps Basko',
    author_email='jazeps.basko@gmail.com',
    maintainer='Jazeps Basko',
    maintainer_email='jazeps.basko@gmail.com',
    license='MIT',
    url='https://github.com/jbasko/wr-profiles',
    description='Switch environment variable based profiles in Python with ease',
    long_description=read('docs/README.md'),
    packages=['wr_profiles'],
    install_requires=[],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
    ],
)



import pytest

from tests.warehouse_profile import WarehouseProfile
from wr_profiles import Property


@pytest.mark.parametrize('name,is_live', [
    ['staging', True],
    ['staging', False],
    [None, True],
    [None, False],
])
def test_sets_profile_property_value(name, is_live):
    profile = WarehouseProfile(name=name, is_live=is_live)

    with pytest.raises(Property.MissingValue):
        _ = profile.username  # noqa

    profile.username = 'the-new-username'
    assert profile.username == 'the-new-username'

    assert 'username' not in profile.__dict__
